@@ -12,6 +12,7 @@ import borg.archive
 import borg.archiver
 import borg.helpers
 import borg.repository
+from dotenv import dotenv_values, load_dotenv
 
 from .options import (
     ArchiveInput,
@@ -22,10 +23,10 @@ from .options import (
     ExclusionOptions,
     ExclusionOutput,
     FilesystemOptions,
-    Options,
+    _Options,
 )
 
-
+# pylint: disable=too-many-public-methods
 class BorgAPI:
     """Automate borg in code"""
 
@@ -44,6 +45,7 @@ class BorgAPI:
         self.archiver.log_json = log_json
         borg.archiver.setup_logging(level=log_level, is_serve=False, json=log_json)
         self.original_stdout = sys.stdout
+        self._previous_dotenv = []
 
     def _run(self, arg_list: List, func: Callable) -> Union[str, dict, None]:
         capture = value = None
@@ -69,7 +71,7 @@ class BorgAPI:
 
         return capture
 
-    def _get_option_list(self, value: dict, options_class: Options) -> List:
+    def _get_option_list(self, value: dict, options_class: _Options) -> List:
         args = self.options | (value or {})
         return options_class(**args).parse()
 
@@ -77,16 +79,70 @@ class BorgAPI:
         self,
         command: str,
         values: dict,
-        options_class: Options,
+        options_class: _Options,
     ) -> List:
         optionals = self.defaults.get(command, {}) | (values or {})
         return options_class(**optionals).parse()
 
+    def set_environ(
+        self,
+        filename: str = None,
+        dictionary: dict = None,
+        **kwargs: Union[bool, str, int],
+    ) -> None:
+        """Load environment variables from file
+
+        If nothing is provided, load_dotenv's default value will be used.
+
+        :param filename: path to environment file, defaults to None
+        :type filename: str, optional
+        :param dictionary: dictionary of environment variables to load, defaults to None
+        :type dictionary: dict, optional
+        :param **kwargs: Environment variables and their values as named args
+        :type **kwargs: Union[bool, str, int]
+        """
+        variables = {}
+        if filename:
+            logging.debug("Loading environment variables from %s", filename)
+            variables = dotenv_values(filename)
+        elif dictionary or kwargs:
+            logging.debug("Loading dictionary with data: %s", variables)
+            variables = dictionary or kwargs
+        else:
+            logging.debug('Looking for ".env" file to load variables from')
+            variables = dotenv_values()
+
+        self._previous_dotenv = variables.keys()
+
+        config = StringIO()
+        for key, value in variables.items():
+            config.write(f"{key}={value}\n")
+        load_dotenv(stream=config)
+
+    def unset_environ(
+        self,
+        *variable: Optional[str],
+    ) -> None:
+        """Remove variables from the environment
+
+        If no variable is provided the values set from the previous call to `set_environ`
+            will be removed.
+
+        :param *variable: variable names to remove
+        :type *variable: Optional[str]
+        """
+        variables = (
+            [k for k in variable if k in os.environ] or
+            [k for k in self._previous_dotenv if k in os.environ]
+        )
+        for var in variables:
+            del os.environ[var]
+
     def init(
         self,
         repository: str,
         encryption: str = "repokey",
-        **options: dict,
+        **options: Union[bool, str, int],
     ) -> Union[str, dict, None]:
         """Initialize an empty repository. A repository is a filesystem directory
         containing the deduplicated data from zero or more archives.
@@ -96,14 +152,14 @@ class BorgAPI:
         :param encryption: select encryption key mode; defaults to "repokey"
         :type encryption: str, optional
         :param **options: optional arguments specific to `init` and common options; defaults to {}
-        :type **options: keyword args
+        :type **options: Union[bool, str, int]
         :return: Stdout of command, None if no output created,
             json dict if json flag used, str otherwise
         :rtype: Union[str, dict, None]
         """
 
         @dataclass
-        class _Optional(Options):
+        class _Optional(_Options):
             append_only: bool = False
             storage_quota: str = None
             make_parent_dirs: bool = False
@@ -124,26 +180,26 @@ class BorgAPI:
     def create(
         self,
         archive: str,
-        *paths: tuple[str],
-        **options: dict,
+        *paths: str,
+        **options: Union[bool, str, int],
     ) -> Union[str, dict, None]:
         """Create a backup archive containing all files found while recursively
         traversing all paths specified.
 
         :param archive: name of archive to create (must be also a valid directory name)
         :type archive: str
-        :param paths: paths to archive
-        :type paths: tuple[str]
+        :param *paths: paths to archive
+        :type *paths: str
         :param **options: optional arguments specific to `create` as well as exclusion,
             filesysem, archive, and common options; defaults to {}
-        :type **options: keyword args
+        :type **options: Union[bool, str, int]
         :return: Stdout of command, None if no output created,
             dict if json flag used, str otherwise
         :rtype: Union[str, dict, None]
         """
 
         @dataclass
-        class _Optional(Options):
+        class _Optional(_Options):
             dry_run: bool = False
             stats: bool = False
             list: bool = False
@@ -175,25 +231,25 @@ class BorgAPI:
     def extract(
         self,
         archive: str,
-        *paths: Optional[tuple[str]],
-        **options: dict,
+        *paths: Optional[str],
+        **options: Union[bool, str, int],
     ) -> Union[str, dict, None]:
         """Extract the contents of an archive.
 
         :param archive: archive to extract
         :type archive: str
-        :param paths: paths to archive
-        :type paths: Optional[tuple[str]]
+        :param *paths: paths to archive
+        :type *paths: Optional[str]
         :param **options: optional arguments specific to `extract` as well as exclusion
             and common options; defaults to {}
-        :type **options: keyword args
+        :type **options: Union[bool, str, int]
         :return: Stdout of command, None if no output created,
             dict if json flag used, str otherwise
         :rtype: Union[str, dict, None]
         """
 
         @dataclass
-        class _Optional(Options):
+        class _Optional(_Options):
             list: bool = False
             dry_run: bool = False
             numeric_owner: bool = False
@@ -219,23 +275,23 @@ class BorgAPI:
 
     def check(
         self,
-        *repository_or_archive: tuple[str],
-        **options: dict,
+        *repository_or_archive: str,
+        **options: Union[bool, str, int],
     ) -> Union[str, dict, None]:
         """Verify the consistency of a repository and the corresponding archives.
 
-        :param repository_or_archive: repository or archive to check consistency of
-        :type repository_or_archive: tuple[str]
+        :param *repository_or_archive: repository or archive to check consistency of
+        :type *repository_or_archive: str
         :param **options: optional arguments specific to `check` as well as archive
             and common options; defaults to {}
-        :type **options: keyword args
+        :type **options: Union[bool, str, int]
         :return: Stdout of command, None if no output created,
             dict if json flag used, str otherwise
         :rtype: Union[str, dict, None]
         """
 
         @dataclass
-        class _Optional(Options):
+        class _Optional(_Options):
             repository_only: bool = False
             archives_only: bool = False
             verify_data: bool = False
@@ -259,7 +315,7 @@ class BorgAPI:
         self,
         archive: str,
         newname: str,
-        **options: dict,
+        **options: Union[bool, str, int],
     ) -> Union[str, dict, None]:
         """Rename an archive in the repository.
 
@@ -269,7 +325,7 @@ class BorgAPI:
         :type newname: str
         :param **options: optional arguments specific to `rename` as well as
             common options; defaults to {}
-        :type **options: keyword args
+        :type **options: Union[bool, str, int]
         :return: Stdout of command, None if no output created,
             dict if json flag used, str otherwise
         :rtype: Union[str, dict, None]
@@ -287,25 +343,25 @@ class BorgAPI:
     def list(
         self,
         repository_or_archive: str,
-        *paths: Optional[tuple[str]],
-        **options: dict,
+        *paths: Optional[str],
+        **options: Union[bool, str, int],
     ) -> Union[str, dict, None]:
         """List the contents of a repository or an archive.
 
         :param repository_or_archive: repository or archive to list contents of
         :type repository_or_archive: str
-        :param paths: paths to list; patterns are supported
-        :type paths: Optional[tuple[str]]
+        :param *paths: paths to list; patterns are supported
+        :type *paths: Optional[str]
         :param **options: optional arguments specific to `list` as well as exclusion,
             archive, and common options; defaults to {}
-        :type **options: keyword args
+        :type **options: Union[bool, str, int]
         :return: Stdout of command, None if no output created,
             dict if json flag used, str otherwise
         :rtype: Union[str, dict, None]
         """
 
         @dataclass
-        class _Optional(Options):
+        class _Optional(_Options):
             short: bool = False
             format: str = None
             json: bool = False
@@ -330,8 +386,8 @@ class BorgAPI:
         self,
         repo_archive_1: str,
         archive_2: str,
-        *paths: Optional[tuple[str]],
-        **options: dict,
+        *paths: Optional[str],
+        **options: Union[bool, str, int],
     ) -> Union[str, dict, None]:
         """Find the differences (file contents, user/group/mode) between archives.
 
@@ -339,18 +395,18 @@ class BorgAPI:
         :type repo_archive_1: str
         :param archive_2: ARCHIVE2 name (no repository location allowed)
         :type archive_2: str
-        :param paths: paths of items inside the archives to compare; patterns are supported
-        :type paths: Optional[tuple[str]]
+        :param *paths: paths of items inside the archives to compare; patterns are supported
+        :type *paths: Optional[str]
         :param **options: optional arguments specific to `diff` as well as exclusion, and
             common options; defaults to {}
-        :type **options: keyword args
+        :type **options: Union[bool, str, int]
         :return: Stdout of command, None if no output created,
             dict if json flag used, str otherwise
         :rtype: Union[str, dict, None]
         """
 
         @dataclass
-        class _Optional(Options):
+        class _Optional(_Options):
             numeric_owner: bool = False
             same_chunker_params: bool = False
             sort: bool = False
@@ -374,25 +430,25 @@ class BorgAPI:
     def delete(
         self,
         repository_or_archive: str,
-        *archives: Optional[tuple[str]],
-        **options: dict,
+        *archives: Optional[str],
+        **options: Union[bool, str, int],
     ) -> Union[str, dict, None]:
         """Delete an archive from the repository or the complete repository
 
         :param repository_or_archive: repository or archive to delete
         :type repository_or_archive: str
-        :param archives: archives to delete
-        :type archives: Optional[tuple[str]]
+        :param *archives: archives to delete
+        :type *archives: Optional[str]
         :param **options: optional arguments specific to `delete` as well as
             archive and common options; defaults to {}
-        :type **options: keyword args
+        :type **options: Union[bool, str, int]
         :return: Stdout of command, None if no output created,
             dict if json flag used, str otherwise
         :rtype: Union[str, dict, None]
         """
 
         @dataclass
-        class _Optional(Options):
+        class _Optional(_Options):
             dry_run: bool = False
             stats: bool = False
             cache_only: bool = False
@@ -416,7 +472,7 @@ class BorgAPI:
     def prune(
         self,
         repository: str,
-        **options: dict,
+        **options: Union[bool, str, int],
     ) -> Union[str, dict, None]:
         """Prune a repository by deleting all archives not matching any of the specified
         retention options.
@@ -425,7 +481,7 @@ class BorgAPI:
         :type repository: str
         :param **options: optional arguments specific to `prune` as well as archive and
             common options; defaults to {}
-        :type **options: keyword args
+        :type **options: Union[bool, str, int]
         :return: Stdout of command, None if no output created,
             dict if json flag used, str otherwise
         :rtype: Union[str, dict, None]
@@ -433,12 +489,13 @@ class BorgAPI:
 
         # pylint: disable=too-many-instance-attributes
         @dataclass
-        class _Optional(Options):
+        class _Optional(_Options):
             dry_run: bool = False
             force: bool = False
             stats: bool = False
             list: bool = False
             keep_within: str = None
+            keep_last: int = None
             keep_secondly: int = None
             keep_minutely: int = None
             keep_hourly: int = None
@@ -464,7 +521,7 @@ class BorgAPI:
     def info(
         self,
         repository_or_archive: str,
-        **options: dict,
+        **options: Union[bool, str, int],
     ) -> Union[str, dict, None]:
         """Display detailed information about the specified archive or repository.
 
@@ -472,14 +529,14 @@ class BorgAPI:
         :type repository_or_archive: str
         :param **options: optional arguments specific to `info` as well as archive and
             common options; defaults to {}
-        :type **options: keyword args
+        :type **options: Union[bool, str, int]
         :return: Stdout of command, None if no output created,
             dict if json flag used, str otherwise
         :rtype: Union[str, dict, None]
         """
 
         @dataclass
-        class _Optional(Options):
+        class _Optional(_Options):
             json: bool = False
 
             # pylint: disable=useless-super-delegation
@@ -499,8 +556,8 @@ class BorgAPI:
         self,
         repository_or_archive: str,
         mountpoint: str,
-        *paths: Optional[tuple[str]],
-        **options: dict,
+        *paths: Optional[str],
+        **options: Union[bool, str, int],
     ) -> Union[str, dict, None]:
         """Mount an archive as a FUSE filesystem.
 
@@ -508,19 +565,21 @@ class BorgAPI:
         :type repository_or_archive: str
         :param mountpoint: where to mount filesystem
         :type mountpoint: str
-        :param paths: paths to extract; patterns are supported
-        :type paths: Optional[tuple[str]]
+        :param *paths: paths to extract; patterns are supported
+        :type *paths: Optional[str]
         :param **options: optional arguments specific to `mount` as well as exclusion,
             archive, and common options; defaults to {}
-        :type **options: keyword args
+        :type **options: Union[bool, str, int]
         :return: Stdout of command, None if no output created,
             dict if json flag used, str otherwise
         :rtype: Union[str, dict, None]
         """
 
+        # pylint: disable=invalid-name
         @dataclass
-        class _Optional(Options):
+        class _Optional(_Options):
             foreground: bool = False
+            o: str = None
 
             # pylint: disable=useless-super-delegation
             def __init__(self, **kwargs):
@@ -541,7 +600,7 @@ class BorgAPI:
     def umount(
         self,
         mountpoint: str,
-        **options: dict,
+        **options: Union[bool, str, int],
     ) -> Union[str, dict, None]:
         """Un-mount a FUSE filesystem that was mounted with `mount`.
 
@@ -549,7 +608,7 @@ class BorgAPI:
         :type mountpoint: str
         :param **options: optional arguments specific to `umount` as well as
             common options; defaults to {}
-        :type **options: keyword args
+        :type **options: Union[bool, str, int]
         :return: Stdout of command, None if no output created,
             dict if json flag used, str otherwise
         :rtype: Union[str, dict, None]
@@ -564,7 +623,7 @@ class BorgAPI:
     def key_change_passphrase(
         self,
         repository: str,
-        **options: dict,
+        **options: Union[bool, str, int],
     ) -> Union[str, dict, None]:
         """Change the passphrase protecting the repository encryption.
 
@@ -572,7 +631,7 @@ class BorgAPI:
         :type repository: str
         :param **options: optional arguments specific to `key change-passphrase` as well as
             common options; defaults to {}
-        :type **options: keyword args
+        :type **options: Union[bool, str, int]
         :return: Stdout of command, None if no output created,
             dict if json flag used, str otherwise
         :rtype: Union[str, dict, None]
@@ -588,7 +647,7 @@ class BorgAPI:
         self,
         repository: str,
         path: str,
-        **options: dict,
+        **options: Union[bool, str, int],
     ) -> Union[str, dict, None]:
         """Copy repository encryption key to another location.
 
@@ -598,14 +657,14 @@ class BorgAPI:
         :type path: str
         :param **options: optional arguments specific to `key export` as well as
             common options; defaults to {}
-        :type **options: keyword args
+        :type **options: Union[bool, str, int]
         :return: Stdout of command, None if no output created,
             dict if json flag used, str otherwise
         :rtype: Union[str, dict, None]
         """
 
         @dataclass
-        class _Optional(Options):
+        class _Optional(_Options):
             paper: bool = False
             qr_html: bool = False
 
@@ -626,7 +685,7 @@ class BorgAPI:
         self,
         repository: str,
         path: str,
-        **options: dict,
+        **options: Union[bool, str, int],
     ) -> Union[str, dict, None]:
         """Restore a key previously backed up with the export command.
 
@@ -636,14 +695,14 @@ class BorgAPI:
         :type path: str
         :param **options: optional arguments specific to `key import` as well as
             common options; defaults to {}
-        :type **options: keyword args
+        :type **options: Union[bool, str, int]
         :return: Stdout of command, None if no output created,
             dict if json flag used, str otherwise
         :rtype: Union[str, dict, None]
         """
 
         @dataclass
-        class _Optional(Options):
+        class _Optional(_Options):
             paper: bool = False
 
             # pylint: disable=useless-super-delegation
@@ -662,7 +721,7 @@ class BorgAPI:
     def upgrade(
         self,
         repository: str,
-        **options: dict,
+        **options: Union[bool, str, int],
     ) -> Union[str, dict, None]:
         """Upgrade an existing, local Borg repository.
 
@@ -670,14 +729,14 @@ class BorgAPI:
         :type repository: str
         :param **options: optional arguments specific to `upgrade` as well as
             common options; defaults to {}
-        :type **options: keyword args
+        :type **options: Union[bool, str, int]
         :return: Stdout of command, None if no output created,
             dict if json flag used, str otherwise
         :rtype: Union[str, dict, None]
         """
 
         @dataclass
-        class _Optional(Options):
+        class _Optional(_Options):
             dry_run: bool = False
             inplace: bool = False
             force: bool = False
@@ -700,8 +759,8 @@ class BorgAPI:
         self,
         archive: str,
         file: str,
-        *paths: Optional[tuple[str]],
-        **options: dict,
+        *paths: Optional[str],
+        **options: Union[bool, str, int],
     ) -> Union[str, dict, None]:
         """Create a tarball from an archive.
 
@@ -709,18 +768,18 @@ class BorgAPI:
         :type archive: str
         :param file: output tar file. “-” to write to stdout instead.
         :type file: str
-        :param paths: paths of items inside the archives to compare; patterns are supported
-        :type paths: Optional[tuple[str]]
+        :param *paths: paths of items inside the archives to compare; patterns are supported
+        :type *paths: Optional[str]
         :param **options: optional arguments specific to `export-tar` as well as exclusion and
             common options; defaults to {}
-        :type **options: keyword args
+        :type **options: Union[bool, str, int]
         :return: Stdout of command, None if no output created,
             dict if json flag used, str otherwise
         :rtype: Union[str, dict, None]
         """
 
         @dataclass
-        class _Optional(Options):
+        class _Optional(_Options):
             tar_filter: str = None
             list: bool = False
 
@@ -739,11 +798,40 @@ class BorgAPI:
 
         return self._run(arg_list, self.archiver.do_export_tar)
 
+    def serve(
+        self,
+        **options: Union[bool, str, int],
+    ) -> Union[str, dict, None]:
+        """Start a repository server process. This command is usually not used manually.
+
+        :return: Stdout of command, None if no output created,
+            dict if json flag used, str otherwise
+        :rtype: Union[str, dict, None]
+        """
+
+        @dataclass
+        class _Optional(_Options):
+            restrict_to_path: str = None
+            restrict_to_repository: str = None
+            append_only: bool = False
+            storage_quota: str = None
+
+            # pylint: disable=useless-super-delegation
+            def __init__(self, **kwargs):
+                super().__init__(**kwargs)
+
+        arg_list = []
+        arg_list.extend(self._get_option_list(options, CommonOptions))
+        arg_list.append("serve")
+        arg_list.extend(self._get_command_list("serve", options, _Optional))
+
+        return self._run(arg_list, self.archiver.do_serve)
+
     def config(
         self,
         repository: str,
         changes: List[tuple[str, str]],
-        **options: dict,
+        **options: Union[bool, str, int],
     ) -> Union[str, dict, None]:
         """Get and set options in a local repository or cache config file.
 
@@ -753,14 +841,14 @@ class BorgAPI:
         :type changes: list[(str,str)]
         :param **options: optional arguments specific to `config` as well as
             common options; defaults to {}
-        :type **options: keyword args
+        :type **options: Union[bool, str, int]
         :return: Stdout of command, None if no output created,
             dict if json flag used, str otherwise
         :rtype: Union[str, dict, None]
         """
 
         @dataclass
-        class _Optional(Options):
+        class _Optional(_Options):
             cache: bool = False
             delete: bool = False
             list: bool = False
@@ -781,3 +869,87 @@ class BorgAPI:
                 return self._run(change, self.archiver.do_config)
         else:
             return self._run(arg_list, self.archiver.do_config)
+
+    def with_lock(
+        self,
+        repository: str,
+        command: str,
+        *args: Union[str, int],
+        **options: Union[bool, str, int]
+    ) -> Union[str, dict, None]:
+        """Run a user-specified command while the repository lock is held.
+
+        :param repository: repository to lock
+        :type repository: str
+        :param command: command to run
+        :type command: str
+        :param *args: command arguments
+        :type *args: Union[str, int]
+        :param **options: optional arguments specific to `config` as well as
+            common options; defaults to {}
+        :type **options: Union[bool, str, int]
+        :return: Stdout of command, None if no output created,
+            dict if json flag used, str otherwise
+        :rtype: Union[str, dict, None]
+        """
+
+        arg_list = []
+        arg_list.extend(self._get_option_list(options, CommonOptions))
+        arg_list.append("with-lock")
+        arg_list.append(repository)
+        arg_list.append(command)
+        arg_list.extend(args)
+
+        return self._run(arg_list, self.archiver.do_with_lock)
+
+    def break_lock(
+        self,
+        repository: str,
+        **options: Union[bool, str, int]
+    ) -> Union[str, dict, None]:
+        """Break the repository and cache locks.
+
+        :param repository: repository for which to break the locks
+        :type repository: str
+        :param **options: optional arguments specific to `config` as well as
+            common options; defaults to {}
+        :type **options: Union[bool, str, int]
+        :return: Stdout of command, None if no output created,
+            dict if json flag used, str otherwise
+        :rtype: Union[str, dict, None]
+        """
+
+        arg_list = []
+        arg_list.extend(self._get_option_list(options, CommonOptions))
+        arg_list.append("break-lock")
+        arg_list.append(repository)
+
+        return self._run(arg_list, self.archiver.do_break_lock)
+
+    def benchmark_crud(
+        self,
+        repository: str,
+        path: str,
+        **options: Union[bool, str, int]
+    ) -> Union[str, dict, None]:
+        """Benchmark borg CRUD (create, read, update, delete) operations.
+
+        :param repository: repository to use for benchmark (must exist)
+        :type repository: str
+        :param path: path were to create benchmark input data
+        :type path: str
+        :param **options: optional arguments specific to `config` as well as
+            common options; defaults to {}
+        :type **options: Union[bool, str, int]
+        :return: Stdout of command, None if no output created,
+            dict if json flag used, str otherwise
+        :rtype: Union[str, dict, None]
+        """
+
+        arg_list = []
+        arg_list.extend(self._get_option_list(options, CommonOptions))
+        arg_list.append("benchmark crud")
+        arg_list.append(repository)
+        arg_list.append(path)
+
+        return self._run(arg_list, self.archiver.do_benchmark_crud)
